
/*
	spiprog.c : CDC-SPI demo
				AVR Programmer Version 0.5

	2010-01-15
	Osamu Tamura @ Recursion Co., Ltd.
*/

#define	_CRT_SECURE_NO_WARNINGS
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>

#define	VendorCode		0x1E	//	Atmel

static const char	*title = "\n\tCDC-SPI AVR Programmer Ver. 0.51 by Osamu Tamura\n";
static const char	*usage = "\tusage: spiprog -c port [-f fusebits(EEHHLL)] [input.hex]";
//	ex. spiprog -c5 -fFEDCBA sample.hex
//	fusebits:	Extended, High, Low Bytes

static HANDLE			hCom;
static unsigned char	xbuf[9];


//	Write and Read n bytes
static int SPI_WR( int n )
{
	DWORD		rwlen;

	WriteFile( hCom, xbuf, n, &rwlen, 0 );
	ReadFile( hCom, xbuf, rwlen, &rwlen, 0 );

	return (int)rwlen==n? 1:0;
}

static int PollRdyBsy( void )
{
	memcpy( xbuf, "\xF0\x00\x00\x00", 4 );
	if( !SPI_WR(4) )
		return -1;

	if( (xbuf[3]&1) ) {
		int		i;

		for( i=0; i<10; i++ ) {
			Sleep(1);
			memcpy( xbuf, "\xF0\x00\x00\x00", 4 );
			if( !SPI_WR(4) )
				return -1;
			if( (xbuf[3]&1)==0 )
				break;
		}
		return xbuf[3]&1;
	}
	else {
		Sleep(30);
		return 0;
	}
}

static int ProgrammingEnable( void )
{
	int			i;

	for( i=0; i<10; i++ ) {
		EscapeCommFunction( hCom, SETBREAK );
		Sleep( 20 );
		EscapeCommFunction( hCom, CLRBREAK );
		Sleep( 30 );
		memcpy( xbuf, "\xAC\x53\x00\x00", 4 );
		if( SPI_WR(4) ) {
			memcpy( xbuf, "\x30\x00\x00\x00", 4 );
			if( SPI_WR(4) && xbuf[1]==0x30 )
				return (int)xbuf[3];
		}
	}
	return 0;
}

static int ReadSignature( unsigned char *sig )
{
	int			addr;

	for( addr=0; addr<3; addr++ ) {
		memcpy( xbuf, "\x30\x00\x00\x00", 4 );
		xbuf[2]	= (unsigned char)addr;
		if( !SPI_WR(4) )
			return -1;
		sig[addr]	= xbuf[3];
	}
	if( sig[2]&0x80 ) {
		sig[1]	= sig[2];
		memcpy( xbuf, "\x30\x00\x04\x00", 4 );
		if( !SPI_WR(4) )
			return -1;
		sig[2]	= xbuf[3];
	}
	return 0;
}

static int ChipErase( void )
{
	memcpy( xbuf, "\xAC\x80\x00\x00", 4 );
	if( !SPI_WR(4) )
		return -1;

	return PollRdyBsy();
}

static int ReadLockBits( void )
{
	memcpy( xbuf, "\x58\x00\x00\x00", 4 );
	if( !SPI_WR(4) )
		return -1;

	return (int)xbuf[3];
}

static int ReadFuseBits( void )
{
	unsigned int	ret = 0;

	memcpy( xbuf, "\x50\x08\x00\x00", 4 );
	if( !SPI_WR(4) )
		return -1;
	ret	= xbuf[3];

	memcpy( xbuf, "\x58\x08\x00\x00", 4 );
	if( !SPI_WR(4) )
		return -1;
	ret	= (ret<<8) | xbuf[3];

	memcpy( xbuf, "\x50\x00\x00\x00", 4 );
	if( !SPI_WR(4) )
		return -1;
	ret	= (ret<<8) | xbuf[3];

	return (int)ret;
}

static int WriteFuseBits( int bits )
{

	if( bits&0xff0000 ) {
		memcpy( xbuf, "\xAC\xA4\x00", 3 );
		xbuf[3]	= (unsigned char)(bits>>16);
		if( !SPI_WR(4) )
			return -1;
		if( PollRdyBsy() )
			return -1;
	}

	memcpy( xbuf, "\xAC\xA8\x00", 3 );
	xbuf[3]	= (unsigned char)(bits>>8);
	if( !SPI_WR(4) )
		return -1;
	if( PollRdyBsy() )
		return -1;

	memcpy( xbuf, "\xAC\xA0\x00", 3 );
	xbuf[3]	= (unsigned char)bits;
	if( !SPI_WR(4) )
		return -1;
	if( PollRdyBsy() )
		return -1;

	return 0;
}

static int ReadProgramMemory( unsigned char *data, size_t size, int page )
{
	unsigned int	addr, mask;
	unsigned char	addrH, addrL;

	mask	= page - 1;
	for( addr=0,size=(size+1)>>1; addr<size; addr++ ) {
		addrH	= (unsigned char)(addr>>8);
		addrL	= (unsigned char)addr;

		xbuf[0]	= 0x20;
		xbuf[1]	= addrH;
		xbuf[2]	= addrL;
		xbuf[4]	= 0x28;
		xbuf[5]	= addrH;
		xbuf[6]	= addrL;
		if( !SPI_WR(8) )
			return -1;
		*data++	= xbuf[3];
		*data++	= xbuf[7];

		if( (addr&mask)==0 )
			putchar( '.' );
	}
	return 0;
}

static int VerifyProgramMemory( unsigned char *data, size_t size, int page )
{
	unsigned int	addr, mask;
	unsigned char	addrH, addrL;

	mask	= page - 1;
	for( addr=0,size=(size+1)>>1; addr<size; addr++ ) {
		addrH	= (unsigned char)(addr>>8);
		addrL	= (unsigned char)addr;

		xbuf[0]	= 0x20;
		xbuf[1]	= addrH;
		xbuf[2]	= addrL;
		xbuf[4]	= 0x28;
		xbuf[5]	= addrH;
		xbuf[6]	= addrL;
		if( !SPI_WR(8) )
			return -1;
		if( *data!=xbuf[3] || *(data+1)!=xbuf[7] ) {
//			printf( "\n\t\tError: [%04X] %04X != %04X\n",
//					addr, (xbuf[3]<<8)|xbuf[7], (*data<<8)|*(data+1) );
			return -2;
		}
		data	+= 2;

		if( (addr&mask)==0 )
			putchar( '.' );
	}
	return 0;
}

static int WriteProgramMemory( unsigned char *data, size_t size, int page )
{
	unsigned int	addr, mask;
	unsigned char	addrL;

	mask	= page - 1;
	page	<<= 1;
	for( addr=0,size=(size+1)>>1; addr<size; addr++ ) {
		addrL	= (unsigned char)(addr & mask);

		//	skip FF block
		if( addrL==0 ) {
			int		i;

			for( i=0; i<page; i++ )
				if( data[i]!=0xff )
					break;
			if( i==page ) {
				addr	+= mask;
				data	+= page;
				continue;
			}
		}

		xbuf[0]	= 0x40;
		xbuf[1]	= 0;
		xbuf[2]	= addrL;
		xbuf[3]	= *data++;
		xbuf[4]	= 0x48;
		xbuf[5]	= 0;
		xbuf[6]	= addrL;
		xbuf[7]	= *data++;
		if( !SPI_WR(8) )
			return -1;

		if( ((addr+1)&mask)==0 ) {
			xbuf[0]	= 0x4C;
			xbuf[1]	= (unsigned char)(addr>>8);
			xbuf[2]	= (unsigned char)addr;
			xbuf[3]	= 0;
			if( !SPI_WR(4) )
				return -1;
			putchar( '.' );
			PollRdyBsy();
		}
	}
	return 0;
}

static void DumpMemory( unsigned char *data, size_t size )
{
	int			i, addr;

	for( i=(int)size-1; i>=0 && data[i]==0xff; i-- )
	;
	if( i<0 )
		i	= 15;
	for( addr=0; addr<=i; addr++ ) {
		if( (addr&15)==0 )
			printf( "\n\t%04X: %02X", addr, data[addr] );
		else
			printf( "%c%02X", (addr&7)? ' ':'-', data[addr] );
	}
	putchar( '\n' );
}

static unsigned int GetHex2( char *data )
{
	unsigned int	val;

	val	= (*data<='9'? *data-'0':*data-'A'+10)<<4;
	data++;
	val	+= *data<='9'? *data-'0':*data-'A'+10;
	return val;
}

#define	MAX_LINE	256

static size_t LoadHex( const char *filename, unsigned char *data, size_t size )
{
	FILE			*fp;
	unsigned int	length, offset, type, segment, addr;
	char			*ptr;
	char			hbuf[MAX_LINE];

	fp	= fopen( filename, "r" );
	if( fp==NULL )
		return 0;

	memset( data, 0xff, size );

	segment	= 0;
	offset	= 0;
	addr	= 0;
	type	= 0;
	while( fgets(hbuf,MAX_LINE,fp)!=NULL ) {
		if( hbuf[0]!=':' )
			continue;
		type	= GetHex2(hbuf+7);
		if( type==0 ) {
			length	= GetHex2(hbuf+1);
			offset	= (GetHex2(hbuf+3)<<8) + GetHex2(hbuf+5);
			addr	= segment + offset;
			for( ptr=hbuf+9; length && addr<size; ptr+=2,length-- )
				data[addr++]	= (unsigned char)GetHex2(ptr);
		}
		else if( type==1 )
			break;
		else if( type==2 )
			segment	= ((GetHex2(hbuf+9)<<8) + GetHex2(hbuf+11)) << 4;
		else if( type==4 )
			segment	= ((GetHex2(hbuf+9)<<8) + GetHex2(hbuf+11)) << 16;
	}
	fclose( fp );

	return type==1? (size_t)addr:(size+1);
}


int main(int argc, char *argv[])
{
	int				i, infile, err;
	int				port, page, lock, fuse, fuse0;
	size_t			size, length;
	unsigned char	*wdata, *rdata;
	unsigned char	signature[3];
	char			device[16];
	char			message[256];

	puts( title );

	hCom	= INVALID_HANDLE_VALUE;
	wdata	= NULL;
	rdata	= NULL;
	port	= -1;
	fuse	= 0;
	length	= 0;
	message[0]	= 0;

	err	= 1;
	for( i=1,infile=0; i<argc; i++ ) {
		if( argv[i][0]!='-' ) {
			infile	= i;
			continue;
		}
		switch( argv[i][1] ) {
			case 'c':
			case 'C':	port = atoi( argv[i]+2 );
						break;
			case 'f':
			case 'F':	fuse = (int)strtol( argv[i]+2, NULL, 16 );
						break;
			default:
						i = argc;	port = -1;
						break;
		}
	}
	if( port<0 ) {
		strcpy( message, usage );
		goto end;
	}

	/*	Open COM device	*/
	err	= 2;
	sprintf( device, "\\\\.\\COM%d", port );
	hCom = CreateFile( device, GENERIC_READ|GENERIC_WRITE,
					0, 0, OPEN_EXISTING, 0, 0 );
	if( hCom==INVALID_HANDLE_VALUE ) {
		sprintf( message, "\tError: COM%d not available.", port );
		goto end;
	}

	/*	Enable programming	*/
	err	= 3;
	if( ProgrammingEnable()!=VendorCode ) {
		sprintf( message, "\tError: Device not found." );
		goto end;
	}

	if( ReadSignature( signature ) )
		goto end;

	/*	Obtain FlashSize and PageSize	*/
	i	= signature[1] & 0x0f;
	size	= 1024 << i;
	page	= 16 << (i>>1);

	printf( "\tSignature:\t%02X-%02X-%02X  (Flash %dKB, Page %dwords)\n",
				signature[0], signature[1], signature[2], 1<<i, page );

	lock	= ReadLockBits();
	printf( "\tLockBits:\t%02X\n", lock );

	fuse0	= ReadFuseBits();
	if( fuse0<0 )
		goto end;
	printf( "\tFuseBits:\t%02X-%02X-%02X\n",
				(fuse0>>16)&0xff, (fuse0>>8)&0xff, fuse0&0xff );
	putchar( '\n' );

	/*	Load hex file	*/
	if( infile ) {
		err	= 4;
		wdata	= (unsigned char *)malloc( size );
		length	= LoadHex( argv[infile], wdata, size );
//		DumpMemory( wdata, length );
		if( length>size )
			goto end;
		if( length )
			printf( "\t%d bytes loaded.\n\n", length );

		/*	Erase Chip	*/
		err	= 5;
		if( ChipErase() )
			goto end;
		printf( "\tDevice erased.\n" );

		err	= 6;
		if( ProgrammingEnable()!=VendorCode )
			goto end;

		/*	Write data	*/
		if( length ) {
			err	= 7;
			printf( "\tWriting " );
			length	= (length+(page<<1)-1) & ~((page<<1)-1);
			if( WriteProgramMemory( wdata, length, page ) )
				goto end;
			printf( " done.\n" );
		}
	}
	else
		length	= size;

	/*	Verify data	*/
	if( length ) {
		if( infile ) {
			err	= 8;
			printf( "\tVerifying " );
			if( VerifyProgramMemory( wdata, length, page ) ) {
				sprintf( message, "\n\t\t\tError: not verified.\n" );
				goto end;
			}
			printf( " done.\n" );
		}
		else if( !fuse ) {
			err	= 9;
			printf( "\tReading " );
			rdata	= (unsigned char *)malloc( length );
			memset( rdata, 0xff, length );
			if( ReadProgramMemory( rdata, length, page ) )
				goto end;
			printf( " done.\n" );
			DumpMemory( rdata, length );
		}
	}

	/*	Write fuse bits	*/
	err	= 10;
	if( fuse ) {
		if( fuse&0xff0000 )
			printf( "\tWriting FuseBits [%02X-%02X-%02X] ...",
					(fuse>>16)&0xff, (fuse>>8)&0xff, fuse&0xff );
		else
			printf( "\tWriting FuseBits [%02X-%02X] ...",
					(fuse>>8)&0xff, fuse&0xff );
		WriteFuseBits( fuse );
		fuse0	= ReadFuseBits();
		if( !(fuse&0xff0000) )
			fuse0	&= 0xffff;
		if( fuse!=fuse0 )
			goto end;
		printf( " done.\n" );
	}

	//	Restart AVR..
	EscapeCommFunction( hCom, SETBREAK );

	err	= 0;
end:
	free( rdata );
	free( wdata );
	CloseHandle( hCom );

	if( message[0]==0 ) {
		if( err )
			sprintf( message, "\n\tError: %d.", err );
		else if( infile && length )
			sprintf( message, "\n\t%d bytes programmed successfully.", length );
	}
	if( message[0] )
		puts( message );

	return err;
}
